/* Pango
 * pangopdflib-layout.c: PDFlib-specific Pango Layout class
 *
 * Copyright (C) 1999 Red Hat Software
 * Copyright (C) 2000 Tor Lillqvist
 * Copyright (C) 2001 Sun Microsystems
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <stdarg.h>
#include <glib.h>

#include <fontconfig/fontconfig.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include "pango-utils.h"
#include "pango-font.h"
#include "pangopdflib.h"
#include "pangopdflib-private.h"
#include "pango-layout.h"
#include "pango-layout-private.h"
#include "pango-context-private.h"
#include "pango.h"

#define PANGO_TYPE_PDFLIB_LAYOUT              (pango_pdflib_layout_get_type ())
#define PANGO_PDFLIB_LAYOUT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), PANGO_TYPE_PDFLIB_LAYOUT, PangoPDFLIBLayout))
#define PANGO_PDFLIB_LAYOUT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), PANGO_TYPE_PDFLIB_LAYOUT, PangoPDFLIBLayoutClass))
#define PANGO_IS_PDFLIB_LAYOUT(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), PANGO_TYPE_PDFLIB_LAYOUT))
#define PANGO_IS_PDFLIB_LAYOUT_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), PANGO_TYPE_PDFLIB_LAYOUT))
#define PANGO_PDFLIB_LAYOUT_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), PANGO_TYPE_PDFLIB_LAYOUT, PangoPDFLIBLayoutClass))

static PangoLayoutClass *parent_class;

struct _PangoPDFLIBLayout
{
  PangoLayout parent_instance;

  gint line_height;
  gint half_leading;
  gint space_before;
  gint space_after;
  gboolean text_altitude_is_custom;
  gint text_altitude;
  gboolean text_depth_is_custom;
  gint text_depth;
  PangoLineStackingStrategy line_stacking_strategy;
};

struct _PangoPDFLIBLayoutClass
{
  PangoLayoutClass parent_class;
};

static void pango_pdflib_layout_class_init (PangoPDFLIBLayoutClass *class);
static void pango_pdflib_layout_init       (PangoPDFLIBLayout      *layout);
static void pango_pdflib_layout_finalize   (GObject         *object);
static void pango_pdflib_layout_check_lines (PangoLayout *layout);
static void pango_pdflib_layout_line_get_extents (PangoLayoutLine *line,
						  PangoRectangle  *ink_rect,
						  PangoRectangle  *logical_rect);

static void pango_pdflib_insert_run (PangoLayoutLine *line,
				     ParaBreakState  *state,
				     const char      *text,
				     PangoItem       *run_item,
				     gboolean         last_run);
/*
static void pango_pdflib_layout_run_get_extents (PangoLayoutRun *run,
						 gboolean       *shape_setp,
						 PangoRectangle *run_ink,
						 PangoRectangle *run_logical);
*/
static void pango_pdflib_calc_half_leading (PangoLayout *layout);
static void pango_pdflib_layout_set_font_description (PangoLayout    *layout,
						      const PangoFontDescription *desc);
static PangoLayout* pango_pdflib_layout_copy (PangoLayout *src);

GType
pango_pdflib_layout_get_type (void)
{
  static GType object_type = 0;

  if (!object_type)
    {
      static const GTypeInfo object_info =
      {
        sizeof (PangoPDFLIBLayoutClass),
        (GBaseInitFunc) NULL,
        (GBaseFinalizeFunc) NULL,
        (GClassInitFunc) pango_pdflib_layout_class_init,
        NULL,           /* class_finalize */
        NULL,           /* class_data */
        sizeof (PangoPDFLIBLayout),
        0,              /* n_preallocs */
        (GInstanceInitFunc) pango_pdflib_layout_init,
      };
      
      object_type = g_type_register_static (PANGO_TYPE_LAYOUT,
                                            "PangoPDFLIBLayout",
                                            &object_info, 0);
    }
  
  return object_type;
}

static void
pango_pdflib_layout_init (PangoPDFLIBLayout *layout)
{
  layout->text_altitude_is_custom = FALSE;
  layout->text_depth_is_custom = FALSE;
}

static void
pango_pdflib_layout_class_init (PangoPDFLIBLayoutClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);
  PANGO_LAYOUT_CLASS (klass)->check_lines = pango_pdflib_layout_check_lines;
  PANGO_LAYOUT_CLASS (klass)->insert_run = pango_pdflib_insert_run;
  PANGO_LAYOUT_CLASS (klass)->set_font_description =
    pango_pdflib_layout_set_font_description;
  PANGO_LAYOUT_CLASS (klass)->get_extents =
    pango_pdflib_layout_line_get_extents;
  PANGO_LAYOUT_CLASS (klass)->copy = pango_pdflib_layout_copy;

  object_class->finalize = pango_pdflib_layout_finalize;
}

static void
pango_pdflib_layout_finalize (GObject *object)
{
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * pango_pdflib_layout_new:
 * @context: a #PangoContext
 * 
 * Create a new #PangoLayout object with attributes initialized to
 * default values for a particular #PangoContext.
 * 
 * Return value: a new #PangoLayout, with a reference count of one.
 **/
PangoLayout *
pango_pdflib_layout_new (PangoContext *context)
{
  PangoPDFLIBLayout *layout;

  g_return_val_if_fail (context != NULL, NULL);

  layout = PANGO_PDFLIB_LAYOUT (g_object_new (pango_pdflib_layout_get_type (),
					      NULL));

  PANGO_LAYOUT (layout)->context = context;
  g_object_ref (G_OBJECT (context));

  return PANGO_LAYOUT (layout);
}

/**
 * pango_pdflib_layout_copy:
 * @src: a #PangoLayout
 * 
 * Does a deep copy-by-value of the @src layout. The attribute list,
 * tab array, and text from the original layout are all copied by
 * value.
 * 
 * Return value: a new #PangoPDFLibLayout identical to @src
 **/
PangoLayout*
pango_pdflib_layout_copy (PangoLayout *src)
{
  PangoLayout *layout;
  
  g_return_val_if_fail (PANGO_IS_PDFLIB_LAYOUT (src), NULL);

  layout = pango_pdflib_layout_new (src->context);

  ((PangoPDFLIBLayout *) layout)->line_height =
    ((PangoPDFLIBLayout *) src)->line_height;
  ((PangoPDFLIBLayout *) layout)->half_leading =
    ((PangoPDFLIBLayout *) src)->half_leading;
  ((PangoPDFLIBLayout *) layout)->space_before =
    ((PangoPDFLIBLayout *) src)->space_before;
  ((PangoPDFLIBLayout *) layout)->space_after =
    ((PangoPDFLIBLayout *) src)->space_after;
  ((PangoPDFLIBLayout *) layout)->text_altitude_is_custom =
    ((PangoPDFLIBLayout *) src)->text_altitude_is_custom;
  ((PangoPDFLIBLayout *) layout)->text_altitude =
    ((PangoPDFLIBLayout *) src)->text_altitude;
  ((PangoPDFLIBLayout *) layout)->text_depth_is_custom =
    ((PangoPDFLIBLayout *) src)->text_depth_is_custom;
  ((PangoPDFLIBLayout *) layout)->text_depth =
    ((PangoPDFLIBLayout *) src)->text_depth;
  ((PangoPDFLIBLayout *) layout)->line_stacking_strategy =
    ((PangoPDFLIBLayout *) src)->line_stacking_strategy;

  PANGO_LAYOUT_CLASS (parent_class)->copy_instance_variables (src,
							      layout);

  return layout;
}

/* Tries to insert as much as possible of the item at the head of
 * state->items onto @line. Three results are possible:
 *
 *  BREAK_NONE_FIT: Couldn't fit anything
 *  BREAK_SOME_FIT: The item was broken in the middle
 *  BREAK_ALL_FIT: Everything fit.
 *  BREAK_EMPTY_FIT: Nothing fit, but that was ok, as we can break at the first char.
 *
 * If @force_fit is TRUE, then BREAK_NONE_FIT will never
 * be returned, a run will be added even if inserting the minimum amount
 * will cause the line to overflow. This is used at the start of a line
 * and until we've found at least some place to break.
 *
 * If @no_break_at_end is TRUE, then BREAK_ALL_FIT will never be
 * returned even everything fits; the run will be broken earlier,
 * or BREAK_NONE_FIT returned. This is used when the end of the
 * run is not a break position.
 */
static BreakResult
pango_pdflib_process_item (PangoLayout     *layout,
			   PangoLayoutLine *line,
			   ParaBreakState  *state,
			   gboolean	    force_fit,
			   gboolean	    no_break_at_end,
			   gint             keep_with_previous_threshold,
			   gint             keep_with_next_threshold,
			   gint             keep_together_threshold)
{
  PangoAttrInt *keep_with_previous_attr = NULL;
  PangoAttrInt *keep_with_next_attr = NULL;
  PangoAttrInt *keep_together_attr = NULL;
  gint keep_with_previous = G_MININT;
  gint keep_with_next = G_MININT;
  gint keep_together = G_MININT;
  PangoRectangle shape_ink;
  PangoRectangle shape_logical;
  PangoItem *item = state->items->data;
  gboolean shape_set = FALSE;
  int width;
  int length;
  int i;
  gboolean processing_new_item = FALSE;

  if (!state->glyphs)
    {
      state->glyphs = pango_glyph_string_new ();
      
      pango_layout_get_item_properties (item, NULL, NULL,
					&shape_ink,
					&shape_logical,
					&shape_set);

      if (shape_set)
	imposed_shape (layout->text + item->offset,
		       item->num_chars,
		       &shape_ink,
		       &shape_logical,
		       state->glyphs);
      else if (layout->text[item->offset] == '\t')
	shape_tab (line, state->glyphs);
      else
	pango_shape (layout->text + item->offset,
		     item->length,
		     &item->analysis,
		     state->glyphs);

      state->log_widths = NULL;
      state->log_widths_offset = 0;

      processing_new_item = TRUE;

      pango_pdflib_layout_get_item_attrs (item,
					  pango_attr_keep_together_within_line_get_type(),
					  &keep_together_attr,
					  pango_attr_keep_with_previous_within_line_get_type(),
					  &keep_with_previous_attr,
					  pango_attr_keep_with_next_within_line_get_type(),
					  &keep_with_next_attr,
					  PANGO_ATTR_INVALID);

      if (keep_with_next_attr != NULL)
	keep_with_next = keep_with_next_attr->value;
      if (keep_with_previous_attr != NULL)
	keep_with_previous = keep_with_previous_attr->value;
      if (keep_together_attr != NULL)
	keep_together = keep_together_attr->value;

#if 0
	g_message ("process_item:: ff: %d; nbae: %d; kwpt: %d; kwp: %d;"
		   " kwnt: %d; kwn: %d; ktt: %d; kt: %d",
		   force_fit,
		   no_break_at_end,
		   keep_with_previous_threshold,
		   keep_with_previous,
		   keep_with_next_threshold,
		   keep_with_next,
		   keep_together_threshold,
		   keep_together);
	{
	gchar *tmp = g_strndup (layout->text + item->offset, item->length);
	g_message ("process_item:: text: %s", tmp);
	g_free (tmp);
	}
#endif
    }

  if (state->remaining_width < 0 &&
      !no_break_at_end &&
      (keep_with_next <= keep_with_next_threshold))	/* Wrapping off */
    {
      insert_run (line, state, layout->text, item, TRUE);
#if 0
      g_message ("process_item:: insert_run: %p", line->runs);
      g_message ("process_item:: neg. rw:: rw: %d", state->remaining_width);
#endif
      return BREAK_ALL_FIT;
    }

  width = 0;
  if (processing_new_item)
    {
      for (i = 0; i < state->glyphs->num_glyphs; i++)
	width += state->glyphs->glyphs[i].geometry.width;
    }
  else
    {
      for (i = 0; i < item->num_chars; i++)
	width += state->log_widths[state->log_widths_offset + i];
    }

  if ((width <= state->remaining_width ||
       (item->num_chars == 1 && !line->runs)) &&
      !no_break_at_end &&
      (keep_with_next <= keep_with_next_threshold))
    {
      state->remaining_width -= width;
      state->remaining_width = MAX (state->remaining_width, 0);
      insert_run (line, state, layout->text, item, TRUE);

#if 0
      g_message ("process_item:: insert_run: %p",
		 line->runs);
      g_message ("process_item:: all fit:: w: %d; rw: %d",
		 width,
		 state->remaining_width);
#endif
      return BREAK_ALL_FIT;
    }
  else if (!force_fit &&
	   keep_together > G_MININT &&
	   keep_together >= keep_together_threshold)
    {
#if 0
      g_message ("process_item:: no insert_run");
      g_message ("process_item:: kt >= ktt:: kt: %d; ktt: %d; w: %d; rw: %d",
		 keep_together,
		 keep_together_threshold,
		 width,
		 state->remaining_width);
#endif
      pango_glyph_string_free (state->glyphs);
      state->glyphs = NULL;
      g_free (state->log_widths);

      return BREAK_EMPTY_FIT;
    }
  else
    {
      int num_chars = item->num_chars;
      int break_num_chars = num_chars;
      int break_width = width;
      int orig_width = width;
      gboolean retrying_with_char_breaks = FALSE;

      if (processing_new_item)
	{
	  state->log_widths = g_new (PangoGlyphUnit, item->num_chars);
	  pango_glyph_string_get_logical_widths (state->glyphs,
						 layout->text + item->offset,
						 item->length,
						 item->analysis.level,
						 state->log_widths);
	}

    retry_break:
      
      /* Shorten the item by one line break
       */
      while (--num_chars >= 0)
	{
	  width -= (state->log_widths)[state->log_widths_offset + num_chars];

	  /* If there are no previous runs we have to take care to grab at least one char. */
	  if (can_break_at (layout,
			    state->start_offset + num_chars,
			    retrying_with_char_breaks) &&
	      (num_chars > 0 || line->runs))
	    {
	      break_num_chars = num_chars;
	      break_width = width;
	      
	      if (width <= state->remaining_width ||
		  (num_chars == 1 && !line->runs))
		break;
	    }
	}

      if (layout->wrap == PANGO_WRAP_WORD_CHAR &&
	  force_fit &&
	  (keep_together > G_MININT &&
	   keep_together >= keep_together_threshold) &&
	  break_width > state->remaining_width &&
	  !retrying_with_char_breaks)
	{
	  retrying_with_char_breaks = TRUE;
	  num_chars = item->num_chars;
	  width = orig_width;
	  break_num_chars = num_chars;
	  break_width = width;
	  goto retry_break;
	}

      if (force_fit ||
	  (break_width <= state->remaining_width))	/* Succesfully broke the item */
	{
	  if (state->remaining_width >= 0)
	    {
	      state->remaining_width -= break_width;
	      state->remaining_width = MAX (state->remaining_width, 0);
	    }
	  
	  if (break_num_chars == item->num_chars)
	    {
	      insert_run (line, state, layout->text, item, TRUE);

	      return BREAK_ALL_FIT;
	    }
	  else if (break_num_chars == 0)
	    {
	      if (keep_with_previous != G_MININT &&
		  keep_with_previous >= keep_with_previous_threshold)
		{
		  /* BREAK_EMPTY_FIT only if don't want to keep with
		   * previous
		   */
		  return BREAK_EMPTY_FIT;
		}
	      else
		{
		  return BREAK_NONE_FIT;
		}
	    }
	  else
	    {
	      PangoItem *new_item;

	      length =
		g_utf8_offset_to_pointer (layout->text + item->offset,
					  break_num_chars) -
		(layout->text + item->offset);

              new_item = pango_item_split (item, length, break_num_chars);
	      
	      insert_run (line, state, layout->text, new_item, FALSE);

	      state->log_widths_offset += break_num_chars;

	      /* Shaped items should never be broken */
	      g_assert (!shape_set);

	      return BREAK_SOME_FIT;
	    }
	}
      else
	{
	  pango_glyph_string_free (state->glyphs);
	  state->glyphs = NULL;
	  g_free (state->log_widths);

	  return BREAK_NONE_FIT;
	}
    }
}

static void
pango_pdflib_process_line (PangoLayout    *layout,
			   ParaBreakState *state)
{
  PangoLayoutLine *line;
  
  gboolean have_break = FALSE;      /* If we've seen a possible break yet */
  int break_remaining_width = 0;    /* Remaining width before adding run with break */
  int break_start_offset = 0;	    /* Start width before adding run with break */
  GSList *break_link = NULL;        /* Link holding run before break */
  gint keep_together_threshold = G_MININT;

  line = pango_layout_line_new (layout);
  line->start_index = state->line_start_index;

  if (state->first_line)
    state->remaining_width =
      (layout->indent >= 0) ? layout->width - layout->indent : layout->width;
  else
    state->remaining_width =
      (layout->indent >= 0) ? layout->width : layout->width + layout->indent;

  while (state->items)
    {
      PangoItem *item = state->items->data;
      BreakResult result;
      int old_num_chars;
      int old_remaining_width;
      gboolean first_item_in_line;
      PangoAttrInt *keep_with_previous_attr = NULL;
      PangoAttrInt *keep_with_next_attr = NULL;
      PangoAttrInt *keep_together_attr = NULL;
      gint keep_with_previous = G_MININT;
      gint keep_with_next = G_MININT;
      gint keep_together = G_MININT;

      pango_pdflib_layout_get_item_attrs (item,
					  pango_attr_keep_together_within_line_get_type(),
					  &keep_together_attr,
					  pango_attr_keep_with_previous_within_line_get_type(),
					  &keep_with_previous_attr,
					  pango_attr_keep_with_next_within_line_get_type(),
					  &keep_with_next_attr,
					  PANGO_ATTR_INVALID);

      if (keep_with_next_attr)
	keep_with_next = keep_with_next_attr->value;
      if (keep_with_previous_attr)
	keep_with_previous = keep_with_previous_attr->value;
      if (keep_together_attr)
	keep_together = keep_together_attr->value;

      keep_together_threshold = MAX (keep_together, keep_together_threshold);

      /*
      if (line->runs && can_break_at (layout, state->start_offset))
	{
	  if (state->remaining_width == 0)
	    goto done;

	  have_break = TRUE;
	  break_at_start = G_MAXINT;
	  break_remaining_width = state->remaining_width;
	  break_start_offset = state->start_offset;
	  break_link = line->runs;
	}
      */

      old_num_chars = item->num_chars;
      old_remaining_width = state->remaining_width;
      first_item_in_line = (line->runs != NULL);
      
      result = pango_pdflib_process_item (layout,
					  line,
					  state,
					  !have_break,
					  FALSE,
					  G_MININT /*break_at_start*/,
					  G_MININT,
					  keep_together_threshold);
#if 0
      g_message ("process_line:: first process_item:: r: %d; rw: %d",
		 result,
		 state->remaining_width);
#endif

      switch (result)
	{
	case BREAK_ALL_FIT:
	  if (can_break_in (layout,
			    state->start_offset,
			    old_num_chars,
			    first_item_in_line))
	    {
	      have_break = TRUE;
	      break_remaining_width = old_remaining_width;
	      break_start_offset = state->start_offset;
	      break_link = line->runs->next;
	    }
	  
	  state->items = g_list_delete_link (state->items, state->items);
	  state->start_offset += old_num_chars;

	  break;

	case BREAK_EMPTY_FIT:
	  goto done;
	  
	case BREAK_SOME_FIT:
	  state->start_offset += old_num_chars - item->num_chars;
	  goto done;
	  
	case BREAK_NONE_FIT:
	  {
	    PangoAttrInt *keep_with_previous_attr = NULL;
	    PangoAttrInt *keep_with_next_attr = NULL;
	    PangoAttrInt *keep_together_attr = NULL;
	    gint keep_with_previous = G_MININT;
	    gint keep_with_next = G_MININT;
	    gint keep_together = G_MININT;

	    /* Back up over unused runs to run where there is a break */
#if 0
	    g_log (G_LOG_DOMAIN,
		   G_LOG_LEVEL_DEBUG,
		   "line->runs: %p; break_link: %p",
		   line->runs,
		   break_link);
#endif
	    while (line->runs && line->runs != break_link)
	      {
		state->items = g_list_prepend (state->items,
					       uninsert_run (line));
#if 0
		g_log (G_LOG_DOMAIN,
		       G_LOG_LEVEL_DEBUG,
		       "line->runs: %p; break_link: %p",
		       line->runs,
		       break_link);
#endif
	      }

	    state->start_offset = break_start_offset;
	    state->remaining_width = break_remaining_width;

	    /* Reshape run to break */
	    item = state->items->data;

	    pango_pdflib_layout_get_item_attrs (item,
						pango_attr_keep_together_within_line_get_type(),
						&keep_together_attr,
						pango_attr_keep_with_previous_within_line_get_type(),
						&keep_with_previous_attr,
						pango_attr_keep_with_next_within_line_get_type(),
						&keep_with_next_attr,
						PANGO_ATTR_INVALID);

	    if (keep_with_next_attr)
	      keep_with_next = keep_with_next_attr->value;
	    if (keep_with_previous_attr)
	      keep_with_previous = keep_with_previous_attr->value;
	    if (keep_together_attr)
	      keep_together = keep_together_attr->value;

	    keep_together_threshold = MAX (keep_together, keep_together_threshold);

#if 0
	    g_message ("process_line:: kwp: %d; kwn: %d; kt: %d; ktt: %d",
		       keep_with_previous,
		       keep_with_next,
		       keep_together,
		       keep_together_threshold);
#endif

	    old_num_chars = item->num_chars;
	    result = pango_pdflib_process_item (layout,
						line,
						state,
						TRUE,
						TRUE,
						keep_with_previous,
						keep_with_next,
						keep_together_threshold);
#if 0
	    g_message ("process_line:: result: %d",
		       result);
#endif

	    g_assert (result == BREAK_SOME_FIT || result == BREAK_EMPTY_FIT);

	    state->start_offset += old_num_chars - item->num_chars; 

	    goto done;
	  }
	}
    }

 done:  
  pango_layout_line_postprocess (line);
  layout->lines = g_slist_prepend (layout->lines, line);
  state->first_line = FALSE;
  state->line_start_index += line->length;
}

static void
pango_pdflib_layout_check_lines (PangoLayout *layout)
{
  const char *start;
  gboolean done = FALSE;
  int start_offset;
  PangoAttrList *attrs;
  PangoAttrIterator *iter;
  
  if (layout->lines)
    return;

  g_assert (!layout->log_attrs);

  /* For simplicity, we make sure at this point that layout->text
   * is non-NULL even if it is zero length
   */
  if (!layout->text)
    pango_layout_set_text (layout, NULL, 0);

  attrs = pango_layout_get_effective_attributes (layout);
  iter = pango_attr_list_get_iterator (attrs);
  
  layout->log_attrs = g_new (PangoLogAttr, layout->n_chars + 1);
  
  start_offset = 0;
  start = layout->text;

  do
    {
      int delim_len;
      const char *end;
      int delimiter_index, next_para_index;
      ParaBreakState state;

      if (layout->single_paragraph)
        {
          delimiter_index = layout->length;
          next_para_index = layout->length;
        }
      else
        {
          pango_find_paragraph_boundary (start,
                                         (layout->text + layout->length) - start,
                                         &delimiter_index,
                                         &next_para_index);
        }

      g_assert (next_para_index >= delimiter_index);
      
      end = start + delimiter_index;
      
      delim_len = next_para_index - delimiter_index;
      
      if (end == (layout->text + layout->length))
	done = TRUE;

      g_assert (end <= (layout->text + layout->length));
      g_assert (start <= (layout->text + layout->length));
      g_assert (delim_len < 4);	/* PS is 3 bytes */
      g_assert (delim_len >= 0);
      
      state.items = pango_itemize (layout->context,
				   layout->text,
				   start - layout->text,
				   end - start,
				   attrs,
				   iter);

      get_items_log_attrs (start, state.items,
                           layout->log_attrs + start_offset,
                           delim_len);

      if (state.items)
	{
	  state.first_line = TRUE;
	  state.start_offset = start_offset;
          state.line_start_index = start - layout->text;
	  
	  state.glyphs = NULL;
	  state.log_widths = NULL;
	  
	  while (state.items)
            pango_pdflib_process_line (layout, &state);
	}
      else
        {
          PangoLayoutLine *empty_line;

          empty_line = pango_layout_line_new (layout);
          empty_line->start_index = start - layout->text; 

          layout->lines = g_slist_prepend (layout->lines,
                                           empty_line);
        }

      if (!done)
        start_offset += g_utf8_strlen (start, (end - start) + delim_len);

      start = end + delim_len;
    }
  while (!done);

  pango_attr_iterator_destroy (iter);
  
  if (attrs != layout->attrs)
    pango_attr_list_unref (attrs);
  
  layout->lines = g_slist_reverse (layout->lines);
}

/**
 * pango_pdflib_layout_set_line_height:
 * @layout:      The #PangoPDFLIBLayout.
 * @line_height: The line-height for the layout.
 * 
 * Sets the line-height of the layout.
 **/
void
pango_pdflib_layout_set_line_height (PangoLayout *layout, gint line_height)
{
  PangoPDFLIBLayout *pdflib_layout = (PangoPDFLIBLayout *) layout;

  g_return_if_fail (layout != NULL);
  g_return_if_fail (PANGO_IS_PDFLIB_LAYOUT (layout));
  g_return_if_fail (line_height >= 0);

  pdflib_layout->line_height = line_height;

  pango_pdflib_calc_half_leading (layout);
}

/**
 * pango_pdflib_layout_set_line_stacking_strategy:
 * @layout:                 The #PangoPDFLIBLayout.
 * @line_stacking_strategy: The line-stacking-strategy for the layout.
 * 
 * Sets the line-stacking-strategy of the layout.
 **/
void
pango_pdflib_layout_set_line_stacking_strategy (PangoLayout *layout,
						PangoLineStackingStrategy line_stacking_strategy)
{
  PangoPDFLIBLayout *pdflib_layout = (PangoPDFLIBLayout *) layout;

  g_return_if_fail (layout != NULL);
  g_return_if_fail (PANGO_IS_PDFLIB_LAYOUT (layout));
  g_return_if_fail (line_stacking_strategy != PANGO_LINE_STACKING_STRATEGY_INVALID);

  pdflib_layout->line_stacking_strategy = line_stacking_strategy;
}

/**
 * pango_pdflib_layout_line_get_extents:
 * @line:     a #PangoLayoutLine
 * @ink_rect: rectangle used to store the extents of the glyph string as drawn
 *            or %NULL to indicate that the result is not needed.
 * @logical_rect: rectangle used to store the logical extents of the glyph string
 *            or %NULL to indicate that the result is not needed.
 * 
 * Compute the logical and ink extents of a layout line. See the documentation
 * for pango_font_get_glyph_extents() for details about the interpretation
 * of the rectangles.
 */
void
pango_pdflib_layout_line_get_extents (PangoLayoutLine *line,
				      PangoRectangle  *ink_rect,
				      PangoRectangle  *logical_rect)
{
  PangoPDFLIBLayout *pdflib_layout = (PangoPDFLIBLayout *) line->layout;
  PangoLayoutLinePrivate *private = (PangoLayoutLinePrivate *) line;
  gint text_altitude = pdflib_layout->text_altitude;
  gint text_depth = pdflib_layout->text_depth;
  gint half_leading = pdflib_layout->half_leading;
  PangoLineStackingStrategy lss = pdflib_layout->line_stacking_strategy;
  GSList *tmp_list;
  int x_pos = 0;
  
  g_return_if_fail (LINE_IS_VALID (line));

  if (!LINE_IS_VALID (line))
    return;

  if (!private->logical_rect)
    {
      private->logical_rect = g_new0 (PangoRectangle, 1);
      private->ink_rect = g_new0 (PangoRectangle, 1);

      private->logical_rect->height = text_altitude - text_depth;
      private->logical_rect->y = -text_altitude;

      tmp_list = line->runs;
      while (tmp_list)
	{
	  PangoLayoutRun *run = tmp_list->data;
	  int new_pos;
	  PangoRectangle run_ink;
	  PangoRectangle run_logical;

	  pango_layout_run_get_extents (run, NULL,
					&run_ink,
					&run_logical);
      
	  new_pos = MIN (private->ink_rect->x, x_pos + run_ink.x);
	  private->ink_rect->width =
	    MAX (private->ink_rect->x + private->ink_rect->width,
		 x_pos + run_ink.x + run_ink.width) - new_pos;
	  private->ink_rect->x = new_pos;
	  
	  new_pos = MIN (private->ink_rect->y, run_ink.y);
	  private->ink_rect->height =
	    MAX (private->ink_rect->y + private->ink_rect->height,
		 run_ink.y + run_ink.height) - new_pos;
	  private->ink_rect->y = new_pos;


	  new_pos = MIN (private->logical_rect->x, x_pos + run_logical.x);
	  private->logical_rect->width =
	    MAX (private->logical_rect->x + private->logical_rect->width,
		 x_pos + run_logical.x + run_logical.width) - new_pos;
	  private->logical_rect->x = new_pos;

	  switch (lss)
	    {
	    case PANGO_LINE_STACKING_STRATEGY_LINE_HEIGHT:
	      new_pos =
		MIN (private->logical_rect->y, run_logical.y - half_leading);
	      private->logical_rect->height =
		MAX (private->logical_rect->y + private->logical_rect->height,
		     run_logical.y + run_logical.height + half_leading) - new_pos;
	      private->logical_rect->y = new_pos;
	      break;
	    case PANGO_LINE_STACKING_STRATEGY_FONT_HEIGHT:
	      private->logical_rect->height = text_altitude - text_depth;
	      private->logical_rect->y = -text_altitude;
	      break;
	    case PANGO_LINE_STACKING_STRATEGY_MAX_HEIGHT:
	      new_pos = MIN (private->logical_rect->y, run_logical.y);
	      private->logical_rect->height =
		MAX (private->logical_rect->y + private->logical_rect->height,
		     run_logical.y + run_logical.height) - new_pos;
	      private->logical_rect->y = new_pos;
	      break;
	    case PANGO_LINE_STACKING_STRATEGY_INVALID:
	    default:
	      break;
	    }
	  x_pos += run_logical.width;
	  tmp_list = tmp_list->next;
	}
  
      if (lss != PANGO_LINE_STACKING_STRATEGY_LINE_HEIGHT)
	{
	  private->logical_rect->y -= half_leading;
	  private->logical_rect->height += 2 * half_leading;
	}

      if (private->logical_rect->height == 0) 
	{
	  PangoRectangle temp_rect;
	  pango_layout_line_get_empty_extents (line, &temp_rect);
	  private->logical_rect->height = temp_rect.height;
	}

    }

  if (logical_rect)
    *logical_rect = *private->logical_rect;
  if (ink_rect)
    *ink_rect = *private->ink_rect;
}

/**
 * pango_pdflib_layout_get_item_attrs:
 * @item:                 #PangoItem which may have extra attributes.
 * @first_attribute_type: #PangoAttrType of a predefined or registered
 *                        attribute type.
 * @...:                  Remainder of list of #PangoAttrType and
 *                        **#PangoAttribute pairs, terminated by
 *                        #PANGO_ATTR_INVALID.
 * 
 * Gets specified attribute types, if present, from the extra
 * attributes of @item.
 *
 * For each specified attribute type, if @item has an attribute of
 * that type, sets the #PangoAttribute pointed to by the following
 * argument to the attribute.
 *
 * If @item does not have an attribute of the specified type, sets the
 * #PangoAttribute pointed to by the following argument to NULL.
 * 
 * Return value: The number of matches found.
 **/
gint
pango_pdflib_layout_get_item_attrs (PangoItem *item,
				    PangoAttrType first_attribute_type,
				    ...)
{
  va_list var_args;
  PangoAttrType arg_type = first_attribute_type;
  gint match_count = 0;

  va_start (var_args, first_attribute_type);

  while (arg_type != PANGO_ATTR_INVALID)
    {
      GSList *tmp_list = item->analysis.extra_attrs;
      PangoAttribute **attr_arg = va_arg (var_args, PangoAttribute**);
      gboolean match = FALSE;

      while (tmp_list)
	{
	  PangoAttribute *item_attr = tmp_list->data;

	  if (item_attr->klass->type == arg_type)
	    {
	      *attr_arg = item_attr;
	      match = TRUE;
	      match_count++;
	      break;
	    }

	  if (!match)
	    *attr_arg = NULL;
	  tmp_list = tmp_list->next;
	}

      arg_type = va_arg (var_args, PangoAttrType);
    }

  return match_count;
}

void
pango_pdflib_insert_run (PangoLayoutLine *line,
			 ParaBreakState  *state,
			 const char      *text,
			 PangoItem       *run_item,
			 gboolean         last_run)
{
  PangoLayoutRunPrivate *private = g_new (PangoLayoutRunPrivate, 1);

  private->run.item = run_item;

  if (last_run && state->log_widths_offset == 0)
    private->run.glyphs = state->glyphs;
  else
    {
      private->run.glyphs = pango_glyph_string_new ();
      
      if (text[run_item->offset] == '\t')
	shape_tab (line, private->run.glyphs);
      else
	pango_shape (text + run_item->offset, run_item->length, &run_item->analysis, private->run.glyphs);
    }

  private->logical_rect = NULL;
  private->ink_rect = NULL;
  private->line = line;
  private->baseline_adjust = 0;

  if (last_run)
    {
      if (state->log_widths_offset > 0)
	pango_glyph_string_free (state->glyphs);
      state->glyphs = NULL;
      g_free (state->log_widths);
    }

  line->runs = g_slist_prepend (line->runs, private);
  line->length += run_item->length;
  /*private->get_extents = pango_pdflib_layout_run_get_extents;*/
  private->get_extents = pango_layout_run_get_extents_real;
}

/* Disabled since baseline table support broken by Pango changing
   to using fontconfig. */
#if 0
static gint
pango_pdflib_baseline_offset_get (PangoPDFLIBBaselineTable *baseline_table,
				  PangoPDFLIBBaseline baseline,
				  gint font_size,
				  gint units_per_EM)
{
  switch (baseline)
    {
    case PANGO_PDFLIB_BASELINE_IDEOGRAPHIC:
      return baseline_table->ideographic * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_ALPHABETIC:
      return baseline_table->alphabetic * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_HANGING:
      return baseline_table->hanging * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_MATHEMATICAL:
      return baseline_table->mathematical * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_CENTRAL:
      return baseline_table->central * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_MIDDLE:
      return baseline_table->middle * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_TEXT_BEFORE_EDGE:
      return baseline_table->text_before_edge * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_TEXT_AFTER_EDGE:
      return baseline_table->text_after_edge * font_size / units_per_EM;
      break;
    case PANGO_PDFLIB_BASELINE_AUTO:
    case PANGO_PDFLIB_BASELINE_USE_SCRIPT:
      /* FIXME: need per-script baseline munging */
    default:
      return 0;
      break;

    }
}

void
pango_pdflib_layout_run_get_extents (PangoLayoutRun *run,
				     gboolean       *shape_setp,
				     PangoRectangle *run_ink,
				     PangoRectangle *run_logical)
{
  PangoLayoutRunPrivate *prun = (PangoLayoutRunPrivate *) run;
  PangoUnderline uline = PANGO_UNDERLINE_NONE;
  int rise = 0;
  PangoRectangle shape_ink;
  PangoRectangle shape_logical;
  gboolean shape_set;
  int match_count = 0;
  PangoAttrInt *line_height_attr = NULL;
  PangoAttrInt *alignment_adjust_attr = NULL;
  PangoAttrInt *alignment_baseline_attr = NULL;
  PangoAttrBaseline *dominant_baseline_attr = NULL;
  PangoAttrInt *db_reset_size_attr = NULL;
  PangoPDFLIBFont *curr_font = (PangoPDFLIBFont *) run->item->analysis.font;
  gint curr_font_size = curr_font->size;
  PangoPDFLIBBaselineTable *curr_baseline_table =
    pango_pdflib_font_baseline_table_get (curr_font);
  gint curr_units_per_EM = curr_baseline_table->units_per_EM;
  gint current_baseline_offset = 0;
  gint line_height = 0;

  if (!prun->logical_rect)
    {
      prun->logical_rect = g_new0 (PangoRectangle, 1);
      prun->ink_rect = g_new0 (PangoRectangle, 1);

      pango_layout_get_item_properties (run->item, &uline, &rise,
					&shape_ink, &shape_logical, &shape_set);

      if (shape_setp)
	*shape_setp = shape_set;

      if (shape_set)
	imposed_extents (run->item->num_chars, &shape_ink, &shape_logical,
			 prun->ink_rect, prun->logical_rect);
      else
	pango_glyph_string_extents (run->glyphs, run->item->analysis.font,
				    prun->ink_rect,
				    prun->logical_rect);

      match_count =
	pango_pdflib_layout_get_item_attrs (run->item,
					    pango_attr_line_height_get_type(),
					    &line_height_attr,
					    pango_attr_alignment_adjust_get_type(),
					    &alignment_adjust_attr,
					    pango_attr_alignment_baseline_get_type(),
					    &alignment_baseline_attr,
					    pango_attr_dominant_baseline_get_type(),
					    &dominant_baseline_attr,
					    pango_attr_db_reset_size_get_type(),
					    &db_reset_size_attr,
					    PANGO_ATTR_INVALID);

      if (dominant_baseline_attr)
	{
	  PangoFontDescription *db_font_desc = dominant_baseline_attr->desc;
	  PangoPDFLIBFont *db_font =
	    (PangoPDFLIBFont *) pango_context_load_font (((PangoLayoutRunPrivate *) run)->line->layout->context, db_font_desc);
	  PangoPDFLIBBaselineTable *db_baseline_table =
	    pango_pdflib_font_baseline_table_get (db_font);
	  gint db_units_per_EM = db_baseline_table->units_per_EM;
	  gint db_font_size = db_font->size;
	  PangoPDFLIBBaseline dominant_baseline =
	    dominant_baseline_attr->baseline;
	  gint dominant_baseline_offset = 0;

	  dominant_baseline_offset = pango_pdflib_baseline_offset_get (db_baseline_table, dominant_baseline, db_font_size, db_units_per_EM);
	  current_baseline_offset = pango_pdflib_baseline_offset_get (curr_baseline_table, dominant_baseline, curr_font_size, curr_units_per_EM);

	  prun->baseline_adjust = dominant_baseline_offset - current_baseline_offset;

#if 0
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "db: %d",
		 dominant_baseline);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "baseline-table: %p",
		 run->item->analysis.font);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "curr font: %s",
		 pango_font_description_to_string (pango_font_describe (run->item->analysis.font)));
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "db font: %s",
		 pango_font_description_to_string (((PangoAttrBaseline *) dominant_baseline_attr)->desc));
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "db /EM: %d",
		 db_units_per_EM);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "db size: %d",
		 db_font_size);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "db offset: %d",
		 dominant_baseline_offset);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "curr /EM: %d",
		 curr_units_per_EM);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "curr size: %d",
		 curr_font_size);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "curr offset: %d",
		 current_baseline_offset);
	  if (db_reset_size_attr)
	    g_log (G_LOG_DOMAIN,
		   G_LOG_LEVEL_DEBUG,
		   "reset_size: %d", db_reset_size_attr->value);
	  g_log (G_LOG_DOMAIN,
		 G_LOG_LEVEL_DEBUG,
		 "****************");
#endif

	  if (alignment_baseline_attr)
	    {
	      PangoPDFLIBBaseline alignment_baseline = alignment_baseline_attr->value;
	      gint alignment_baseline_db_offset = 0;
	      gint alignment_baseline_curr_offset = 0;

	      alignment_baseline_db_offset = pango_pdflib_baseline_offset_get (db_baseline_table, alignment_baseline, db_font_size, db_units_per_EM);
	      alignment_baseline_curr_offset = pango_pdflib_baseline_offset_get (curr_baseline_table, alignment_baseline, curr_font_size, curr_units_per_EM);

	      prun->baseline_adjust += alignment_baseline_db_offset - alignment_baseline_curr_offset;
	      /*
	      g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "ab: %d", alignment_baseline);
	      g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "ab curr offset: %d", alignment_baseline_curr_offset);
	      g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "ab db offset: %d", alignment_baseline_db_offset);
	      */
	    }

	  if (alignment_adjust_attr)
	    {
	      PangoPDFLIBBaseline alignment_adjust = alignment_adjust_attr->value;
	      gint alignment_adjust_curr_offset = 0;

	      alignment_adjust_curr_offset = pango_pdflib_baseline_offset_get (curr_baseline_table, alignment_adjust, curr_font_size, curr_units_per_EM);

	      prun->baseline_adjust -= alignment_adjust_curr_offset;
	      /*
	      g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "aa: %d", alignment_adjust);
	      g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "aa curr offset: %d", alignment_adjust_curr_offset);
	      */
	    }
	}
      /*
      g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "adjust: %d", prun->baseline_adjust);
      */

      if (line_height_attr)
	{
	  line_height = line_height_attr->value;
	  /*
	  g_log (G_LOG_DOMAIN,
	     G_LOG_LEVEL_DEBUG,
	     "line-height: %d", line_height);
	  */
	  prun->ink_rect->height = line_height;
	  prun->logical_rect->height = line_height;

	}
      else
	{
	  switch (uline)
	    {
	    case PANGO_UNDERLINE_NONE:
	      break;
	    case PANGO_UNDERLINE_SINGLE:
	      prun->ink_rect->height = MAX (prun->ink_rect->height, 2 * PANGO_SCALE - prun->ink_rect->y);
	      prun->logical_rect->height = MAX (prun->logical_rect->height, 2 * PANGO_SCALE - prun->logical_rect->y);
	      break;
	    case PANGO_UNDERLINE_DOUBLE:
	      prun->ink_rect->height = MAX (prun->ink_rect->height, 4 * PANGO_SCALE - prun->ink_rect->y);
	      prun->logical_rect->height = MAX (prun->logical_rect->height, 4 * PANGO_SCALE - prun->logical_rect->y);
	      break;
	    case PANGO_UNDERLINE_LOW:
	      /* FIXME: Should this simply be prun->logical_rect->height += 2 * PANGO_SCALE instead?
	       */
	      prun->ink_rect->height += 2 * PANGO_SCALE;
	      prun->logical_rect->height = MAX (prun->logical_rect->height,
					       prun->ink_rect->y + prun->ink_rect->height + 2 * PANGO_SCALE - prun->logical_rect->y);
	      break;
	    }
	}
      /*
      g_message ("******************");
      g_message ("run:: ink:: y: %d; height: %d", prun->ink_rect->y, prun->ink_rect->height);
      g_message ("run:: log:: y: %d; height: %d", prun->logical_rect->y, prun->logical_rect->height);
      g_message ("*********");
      */
      prun->ink_rect->y -= prun->baseline_adjust;
      prun->logical_rect->y -= prun->baseline_adjust;

      if (rise != 0)
	{
	  prun->ink_rect->y -= rise;
	  prun->logical_rect->y -= rise;
	}
      /*
      g_message ("run:: ink:: y: %d; height: %d", prun->ink_rect->y, prun->ink_rect->height);
      g_message ("run:: log:: y: %d; height: %d", prun->logical_rect->y, prun->logical_rect->height);
      */
    }

  if (run_logical)
    *run_logical = *prun->logical_rect;
  if (run_ink)
    *run_ink = *prun->ink_rect;
}
#endif

void
pango_pdflib_calc_half_leading (PangoLayout *layout)
{
  PangoPDFLIBLayout *pdflib_layout = (PangoPDFLIBLayout *) layout;
  /*
  PangoPDFLIBFont *pdflib_font;
  PangoPDFLIBBaselineTable *baseline_table = NULL;
  gint font_size;
  gint line_height;
  */

  if (!layout->font_desc || !pdflib_layout->line_height)
    {
      pdflib_layout->half_leading = 0;
      return;
    }

  return;
  /*
  font_size = pango_font_description_get_size (layout->font_desc);
  line_height = pdflib_layout->line_height;
  pdflib_font =
    (PangoPDFLIBFont *) pango_context_load_font (layout->context, layout->font_desc);

  baseline_table =
    pango_pdflib_font_baseline_table_get (pdflib_font);

  pdflib_layout->text_altitude = baseline_table->text_before_edge * font_size / baseline_table->units_per_EM;
  pdflib_layout->text_depth = baseline_table->text_after_edge * font_size / baseline_table->units_per_EM;

  pdflib_layout->half_leading = (line_height - (pdflib_layout->text_altitude - pdflib_layout->text_depth)) / 2;
  */
  /*
  g_message ("half-leading:: lh: %d; ta: %d; td: %d; u/E: %d; fs: %d", line_height, pdflib_layout->text_altitude, pdflib_layout->text_depth, baseline_table->units_per_EM, font_size);
  g_message ("half-leading:: %d", pdflib_layout->half_leading);
  */
}

/**
 * pango_layout_set_font_description:
 * @layout: a #PangoLayout
 * @desc: the new pango font description, or %NULL to unset the
 *        current font description.
 * 
 * Set the default font description for the layout. If no font 
 * description is set on the layout, the font description from
 * the layout's context is used.
 **/
void
pango_pdflib_layout_set_font_description (PangoLayout                *layout,
					  const PangoFontDescription *desc)
{
  g_return_if_fail (layout != NULL);
  g_return_if_fail (PANGO_IS_PDFLIB_LAYOUT (layout));

  pango_layout_set_font_description_real (layout, desc);
  pango_pdflib_calc_half_leading (layout);
}
